/*
IOT Sensor Board SenseBox Demo with configuration webpage
for ESP32 Lolin
Elektor / Jens Nickel
My Journey in the IoT
*/



#include <WiFi.h>

#include <Preferences.h>   // this library is used to get access to Non-volatile storage (NVS) of ESP32
// see https://github.com/espressif/arduino-esp32/blob/master/libraries/Preferences/examples/StartCounter/StartCounter.ino



// +++++++++++++++++++ Start of Display Library +++++++++++++++++++


/**
 * The MIT License (MIT)
 *
 * Copyright (c) 2016 by Daniel Eichhorn
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 * based on sketch created by JohnnyFRX 17Jun17
 */


#include "SSD1306.h" // alias for `#include "SSD1306Wire.h"'

// Initialize the OLED display using Wire library, 5 and 4 are the I2C pins
SSD1306  display(0x3c, 5, 4);


void DisplaySetup()
{
  display.init();
  display.flipScreenVertically();
  display.setFont(ArialMT_Plain_16); 
}


// print text to display, use \n in string for line break
void DisplayPrint(String DisplayText) 
{
  display.clear();
 
  display.setColor(WHITE);
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  display.drawString(64, 5, DisplayText);
  display.setFont(ArialMT_Plain_16);

  display.display();
   
}




// +++++++++++++++++++ End of Display library +++++++++++++++++++++



// +++++++++++++++++++ Start of Webserver Library +++++++++++++++++++

// as ESP32 Arduino Version

WiFiClient myclient;
WiFiServer server(80);


void Webserver_Start()
{
  server.begin();     // Start TCP/IP-Server on ESP32
}



String WebRequestHostAddress;     // global variable used to store Server IP-Address of HTTP-Request


//  Call this function regularly to look for client requests
//  template see https://github.com/espressif/arduino-esp32/blob/master/libraries/WiFi/examples/SimpleWiFiServer/SimpleWiFiServer.ino
//  returns empty string if no request from any client
//  returns GET Parameter: everything after the "/?" if ADDRESS/?xxxx was entered by the user in the webbrowser
//  returns "-" if ADDRESS but no GET Parameter was entered by the user in the webbrowser
//  remark: client connection stays open after return
String Webserver_GetRequestGETParameter()
{
  String GETParameter = "";
  
  myclient = server.available();   // listen for incoming clients

  //Serial.print(".");
  
  if (myclient) {                            // if you get a client,
    Serial.println("New Client.");           // print a message out the serial port
    String currentLine = "";                 // make a String to hold incoming data from the client
    
    while (myclient.connected()) {           // loop while the client's connected
      
      if (myclient.available()) {            // if there's bytes to read from the client,
        
        char c = myclient.read();            // read a byte, then
        Serial.write(c);                     // print it out the serial monitor

        if (c == '\n') {                     // if the byte is a newline character

          // if the current line is blank, you got two newline characters in a row.
          // that's the end of the client HTTP request
          if (currentLine.length() == 0) {
            
            if (GETParameter == "") {GETParameter = "-";};    // if no "GET /?" was found so far in the request bytes, return "-"
            
            // break out of the while loop:
            break;
        
          } else {    // if you got a newline, then clear currentLine:
            currentLine = "";
          }
          
        } else if (c != '\r') {  // if you got anything else but a carriage return character,
          currentLine += c;      // add it to the end of the currentLine
        }

        if (c=='\r' && currentLine.startsWith("GET /?")) 
        // we see a "GET /?" in the HTTP data of the client request
        // user entered ADDRESS/?xxxx in webbrowser, xxxx = GET Parameter
        {
          
          GETParameter = currentLine.substring(currentLine.indexOf('?') + 1, currentLine.indexOf(' ', 6));    // extract everything behind the ? and before a space
                    
        }


        if (c=='\r' && currentLine.startsWith("Host:")) 
        // we see a "Host:" in the HTTP data of the client request
        // user entered ADDRESS or ADDRESS/... in webbrowser, ADDRESS = Server IP-Address of HTTP-Request
        {
          int IndexOfBlank = currentLine.indexOf(' ');
          WebRequestHostAddress = currentLine.substring(IndexOfBlank + 1, currentLine.length());    // extract everything behind the space character and store Server IP-Address of HTTP-Request
          
        }
        
      }
      
    }
    
  }

  return GETParameter;
}



// Send HTML page to client, as HTTP response
// client connection must be open (call Webserver_GetRequestGETParameter() first)
void Webserver_SendHTMLPage(String HTMLPage)
{
   String httpResponse = "";

   // begin with HTTP response header
   httpResponse += "HTTP/1.1 200 OK\r\n";
   httpResponse += "Content-type:text/html\r\n\r\n";

   // then the HTML page
   httpResponse += HTMLPage;

   // The HTTP response ends with a blank line:
   httpResponse += "\r\n";
    
   // send it out to TCP/IP client = webbrowser 
   myclient.println(httpResponse);

   // close the connection
   myclient.stop();
    
   Serial.println("Client Disconnected.");
   
};



// +++++++++++++++++++ End of Webserver library +++++++++++++++++++++






// this section handles configuration values which can be configured via webpage form in a webbrowser

// 8 configuration values max
String ConfigName[8];     // name of the configuration value
String ConfigValue[8];    // the value itself (String)
int    ConfigStatus[8];   // status of the value    0 = not set    1 = valid   -1 = not valid
int    ConfigType[8];     // type of the value    0 = not set    1 = String (textbox)   2 = Byte (slider)


// Initalize the values 
void InitializeConfigValues()
{
  for (int count = 0; count < 8; count++)
  {
    ConfigName[count] = "";
    ConfigValue[count] = "";
    ConfigStatus[count] = 0;
    ConfigType[count] = 0;
  }
}


// Build a HTML page with a form which shows textboxes to enter the values
// returns the HTML code of the page
String EncodeFormHTMLFromValues(String TitleOfForm, int CountOfConfigValues)
{
   // Head of the HTML page
   String HTMLPage = "<!DOCTYPE html><html><body><h2>" + TitleOfForm + "</h2><form><table>";

   // for each configuration value
   for (int c = 0; c < CountOfConfigValues; c++)
   {
      // set background color by the status of the configuration value
      String StyleHTML = "";
      if (ConfigStatus[c] == 0) { StyleHTML = " Style =\"background-color: #FFE4B5;\" " ;};   // yellow
      if (ConfigStatus[c] == 1) { StyleHTML = " Style =\"background-color: #98FB98;\" " ;};   // green
      if (ConfigStatus[c] == -1) { StyleHTML = " Style =\"background-color: #FA8072;\" " ;};  // red

      String TableRowHTML = "";
  
      if (ConfigType[c] == 1)    //  string value, entered in text box
      {
        // build the HTML code for a table row with configuration value name and the value itself inside a textbox   
        TableRowHTML = "<tr><th>" + ConfigName[c] + "</th><th><input name=\"" + ConfigName[c] + "\" value=\"" + ConfigValue[c] + "\" " + StyleHTML + " /></th></tr>";
      }
  
      if (ConfigType[c] == 2)
      {
        // build the HTML code for a table row with configuration value name and the byte value represented by a slider 
        TableRowHTML = "<tr><th>" + ConfigName[c] + "</th><th><input type=\"range\" min=\"1\"  max=\"255\" name=\"" + ConfigName[c] + "\" value=\"" + ConfigValue[c] + "\" " + StyleHTML + " /></th></tr>";
      }
      
      // add the table row HTML code to the page
      HTMLPage += TableRowHTML;
   }

   // add the submit button
   HTMLPage += "</table><br/><input type=\"submit\" value=\"Submit\" />";

   // footer of the webpage
   HTMLPage += "</form></body></html>";
   
   return HTMLPage;
}


// decodes Espace-Codes in URL strings
String DecodeURLChars(String strURL)
{
  strURL.replace("%24", "$");
  strURL.replace("%2F", "/");
  strURL.replace("%3A", ":");
  strURL.replace("%23", "#");
  strURL.replace("%40", "@");
  strURL.replace("%7B", "{");
  strURL.replace("%7D", "}");
  strURL.replace("%22", String(char('\"')));
  return strURL;
}


// Decodes a GET parameter (expression after ? in URI (URI = expression entered in address field of webbrowser)), like "Country=Germany&City=Aachen"
// and set the ConfigValues
int DecodeGETParameterAndSetConfigValues(String GETParameter)
{
   
   int posFirstCharToSearch = 1;
   int count = 0;
   
   // while a "&" is in the expression, after a start position to search
   while (GETParameter.indexOf('&', posFirstCharToSearch) > -1)
   {
     int posOfSeparatorChar = GETParameter.indexOf('&', posFirstCharToSearch);  // position of & after start position
     int posOfValueChar = GETParameter.indexOf('=', posFirstCharToSearch);      // position of = after start position
  
     String strChars = GETParameter.substring(posOfValueChar + 1, posOfSeparatorChar);  // extract everything between = and & and enter it in the ConfigValue
     
     ConfigValue[count] = DecodeURLChars(strChars);  //

     posFirstCharToSearch = posOfSeparatorChar + 1;  // shift the start position to search after the &-char 
     count++;
   }

   // no more & chars found
   
   int posOfValueChar = GETParameter.indexOf('=', posFirstCharToSearch);       // search for =
   
   String strChars = GETParameter.substring(posOfValueChar + 1, GETParameter.length());  // extract everything between = and end of string
   
   ConfigValue[count] = DecodeURLChars(strChars);  // 
   
   count++;

   return count;  // number of values found in GET parameter
}








// +++++++++++++++++++ Start of Application Code +++++++++++++++++++




// include WiFi Network library
#include <WiFiNetwork.h>

// construct the object WiFiNetwork of class WiFiNetwork
WiFiNetwork myWiFiNetwork = WiFiNetwork();


// include TCP Client library
#include <TCPClient.h>

// construct the object mqttTCPClient of class TCPClient
TCPClient mqttTCPClient = TCPClient();


// include MQTT Client for TLS library
#include <MQTTClient.h>

// construct the object myMQTTClient of class MQTTClient
MQTTClient myMQTTClient = MQTTClient();

Preferences preferences;   // we must generate this object of the preference library

// variable to hold status of router network login
byte RouterNetworkDeviceState; 
#define NETWORKSTATE_LOGINREQUESTED       2
#define NETWORKSTATE_LOGGED               1
#define NETWORKSTATE_NOTLOGGED            0

// variable to hold status of TCP and MQTT connection to MQTT broker
byte TCPClient_Connected;
byte MQTTClient_Connected;

// variable to hold last MQTT command
byte MQTTClient_LastCommand;
#define MQTTCONNECT       16
#define MQTTPUBLISH       48
#define MQTTSUBSCRIBE     128
#define MQTTPINGREQ       192

// count the published messages
int MQTTClient_PublishCount;

// we need these variables for MQTT and Ping timing
byte CounterForMQTT;
byte CounterForPing;

// for timing of next attempt to login in router network
byte CounterForLoginRouterNetworkAttempt;

// last value of ADC sampling
int SENSOR_LastValue;



#define PIN_SENSOR     39   // the analog input for sensor



// +++++++++++++++++++ Start of GUI functions +++++++++++++++++++++
// implementation is hardware dependent


// some support functions for display
String DisplayText = "";
byte HeartbeatCount = 0;

void UpdateDisplay()    // will be called on every heartbeat and updated status
{
    DisplayPrint(String(HeartbeatCount & 1) + " " + DisplayText);    // 0..7 as heartbeat sign + the status of the device
}


// Setup of the GUI
// call this function on startup
void GUI_Setup()
{
  DisplaySetup();
  DisplayPrint("ESP32");  
}


// call this function to show a heartbeat sign 
void GUI_Heartbeat()
{
    HeartbeatCount++;  // 0..255 again and again
    UpdateDisplay();
}


// call this function to show updated status of the device
void GUI_Update()
{
   String Line0 = "";
   String Line1 = "";
   String Line2 = "";
   
   // first line for the display = connection status
   if (RouterNetworkDeviceState == NETWORKSTATE_LOGGED)    // logged in network               
   {
         if ((TCPClient_Connected == true) && (MQTTClient_Connected == false))                
         {
            Line0 ="NO MQTT";         // connection with TCP, but not MQTT    
         }
         
         if (TCPClient_Connected == false)                
         {
            Line0 ="NO TCP";         // TCP not connected
         }
      
         if (MQTTClient_Connected == true)                
         {
            Line0 = "MQTT";           // MQTT connected = normal mode of the device  
         }
   }
   
   if (RouterNetworkDeviceState == NETWORKSTATE_NOTLOGGED)     // not logged in network     
   {
      Line0 = "NO NETWORK";     
   }

   if (RouterNetworkDeviceState == NETWORKSTATE_LOGINREQUESTED)   // login requested, if user submits new config values
   {
      Line0 = "NETW REQ";     
   }

   // second line for the display = MQTT command
   if (MQTTClient_LastCommand == MQTTCONNECT)
   {
      Line1 = "CONNECT"; 
   }

   if (MQTTClient_LastCommand == MQTTPUBLISH)
   {      
      Line1 = "PUB "; 
      Line1 += String(MQTTClient_PublishCount); 
   }

   if (MQTTClient_LastCommand == MQTTPINGREQ)
   {
      
      Line1 = "PING"; 
   }

   // third line for the display = sensor value
   Line2 = String(SENSOR_LastValue);

   // concatenate 3 display lines
   DisplayText = Line0 + "\n" + Line1 + "\n" + Line2;

   // and show it on the display
   UpdateDisplay();
   
}



void setup() 
{
  
  myMQTTClient.myTCPClient = mqttTCPClient;    // tell the MQTT object the TCP object because MQTT library is using the TCP/IP library too
    
  GUI_Setup();

  // Open serial communications and wait for port to open
  Serial.begin(19200);
  Serial.println("Hello ESP32!");

  // put the ESP32 in both modes
  myWiFiNetwork.SetBothModesNetworkStationAndAccessPoint();

  // Start access point with SSID "ESP32_MyJourneyIoT"
  int success2 = myWiFiNetwork.AccessPointStart("ESP32_MyJourneyIoT");

  Serial.println(myWiFiNetwork.GetIPAddressInOwnNetwork());

  // start the webserver to listen for request of clients (in LAN or own ESP32 network)
  Webserver_Start();


  // initialize config values and set names
  InitializeConfigValues();
  ConfigName[0] = "LED";
  ConfigType[0] = 1;     // type textbox
  
  ConfigName[1] = "SSID";
  ConfigType[1] = 1;
  
  ConfigName[2] = "PASSWORD";
  ConfigType[2] = 1;
  
  ConfigName[3] = "BROKER ADDRESS";
  ConfigType[3] = 1;

  ConfigName[4] = "CLIENT/DEVICE ID";
  ConfigType[4] = 1; 

  ConfigName[5] = "SENSOR ID/TOKEN";
  ConfigType[5] = 1;         

  ConfigName[6] = "TOPIC";
  ConfigType[6] = 1;         

  ConfigName[7] = "PAYLOAD";
  ConfigType[7] = 1;         


  preferences.begin("MyJourney", false);   // see https://github.com/espressif/arduino-esp32/blob/master/libraries/Preferences/examples/StartCounter/StartCounter.ino 

  // takeout 7 Strings out of the Non-volatile storage
  String strSSID = preferences.getString("SSID", "");
  String strPassword = preferences.getString("Password", "");
  String strBrokerAddress = preferences.getString("BrokerAddress", "");
  String strDeviceID = preferences.getString("DeviceID", "");
  String strSensorID = preferences.getString("SensorID", "");
  String strTopic = preferences.getString("Topic", "");
  String strPayload = preferences.getString("Payload", "");
  
  // put the NVS stored values in RAM for the program
  ConfigValue[1] = strSSID;
  ConfigValue[2] = strPassword;
  ConfigValue[3] = strBrokerAddress;
  ConfigValue[4] = strDeviceID;
  ConfigValue[5] = strSensorID;
  ConfigValue[6] = strTopic;
  ConfigValue[7] = strPayload;
 
  // initialize state variables and counters
  MQTTClient_Connected = false;
  TCPClient_Connected = false;

  MQTTClient_PublishCount = 0;

  CounterForMQTT = 0;
  CounterForPing = 0;
  
  RouterNetworkDeviceState = NETWORKSTATE_NOTLOGGED;  
  CounterForLoginRouterNetworkAttempt = 40;    

  GUI_Update();
}



// check the ConfigValues and set ConfigStatus
// process the first ConfigValue to switch something
void ProcessAndValidateConfigValues(int countValues)
{
  if (countValues > 8) {countValues = 8;};

  // store the second to eigth value in non-volatile storage
  if (countValues > 4)
  {
    preferences.putString("SSID",          ConfigValue[1]);
    preferences.putString("Password",      ConfigValue[2]);
    preferences.putString("BrokerAddress", ConfigValue[3]);
    preferences.putString("DeviceID",      ConfigValue[4]);
    preferences.putString("SensorID",      ConfigValue[5]);
    preferences.putString("Topic",         ConfigValue[6]);
    preferences.putString("Payload",       ConfigValue[7]);
  }

}



// Read incoming request from client (if there is any)
// process values in GET parameter
// and send out new HTML page to client
// returns 1 if GET parameter with new values was received, after URL address 192.168.4.1 (as a sign that user wants to configure device once new)
// returns 0 if any other case 
int GetConfigRequestAndSendHTMLPage()
{
    int ConfigRequestReceived = 0;
    
    String GETParameter = Webserver_GetRequestGETParameter();   // look for client request
  
    if (GETParameter.length() > 0)        // we got a request, client connection stays open
    {
      if (GETParameter.length() > 1)      // request contains some GET parameter
      {
          
          int countValues = DecodeGETParameterAndSetConfigValues(GETParameter);     // decode the GET parameter and set ConfigValues
  
          if (WebRequestHostAddress == "192.168.4.1")                               // the user entered this address in browser, with GET parameter values for configuration
          {
    
                ProcessAndValidateConfigValues(8);                                  // check and process 8 ConfigValues, store SSID, Password, DeviceID and so on in non-volatile storage
            
                ConfigRequestReceived = 1;
          }
          
      }

      String HTMLPage;
       
      if (WebRequestHostAddress == "192.168.4.1")                  //   the user entered this address in the browser, to get the configuration webpage               
      {
          HTMLPage = EncodeFormHTMLFromValues("ESP32 Webserver CONFIG", 8) + "<br>IP Address: " + myWiFiNetwork.GetOwnIPAddressInRouterNetwork();       // build a new Configuration webpage with form and new ConfigValues entered in textboxes or represented by sliders
      }
      
      Webserver_SendHTMLPage(HTMLPage);    // send out the webpage to client = webbrowser and close client connection
      
  }

  return ConfigRequestReceived;
}




// try to login to router network
void LoginRouterNetwork()
{
  String strSSID = preferences.getString("SSID", "");                 // takeout SSID and Password out of non-volatile storage  
  String strPassword = preferences.getString("Password", "");              

  char* txtSSID = const_cast<char*>(strSSID.c_str());                 // convert to char*
  char* txtPassword = const_cast<char*>(strPassword.c_str());         // https://coderwall.com/p/zfmwsg/arduino-string-to-char

  int successDisconnect = myWiFiNetwork.RouterNetworkDisconnect();             // disconnect from router network

  int successConnect = myWiFiNetwork.RouterNetworkConnect(txtSSID, txtPassword);   // then try to connect once new with new login-data

  // print out local address of ESP32 in Router network (LAN)
  Serial.println(myWiFiNetwork.GetOwnIPAddressInRouterNetwork());
    
  if (successConnect == 1)
  {      
      RouterNetworkDeviceState = NETWORKSTATE_LOGGED;                 // set the state
      CounterForMQTT = 0;
  }
  else
  {
      RouterNetworkDeviceState = NETWORKSTATE_NOTLOGGED;              // set the state             
  }   

  // we are just logged in in network, so no MQTT connection
  TCPClient_Connected = false;
  MQTTClient_Connected = false;
  
}




// Connect to MQTT Broker
int ConnectToMQTTBroker()
{   
   
    int MQTTConnectingSuccessful = false;
    
    // closes TCP connection
    mqttTCPClient.Close();

    // get the broker address out of non-volatile storage
    String strBrokerAddress = preferences.getString("BrokerAddress", "");
    char* txtBrokerAddress = const_cast<char*>(strBrokerAddress.c_str());                 // convert to char*
       
    // TCP-connect to MQTT Broker TCP Server
    int TCP_Connect_Result = mqttTCPClient.Connect(txtBrokerAddress, 1883);
    
    // Success?
    if (TCP_Connect_Result != 1)
    {  
      TCPClient_Connected = false;
    }
    else
    {
      TCPClient_Connected = true;

      // take Device Name out of non-volatile storage
      String strDeviceID = preferences.getString("DeviceID", "");
      
      // MQTT CONNECT to MQTT broker , use Device Name of your device as Client ID
      int MQTT_Connect_Result = myMQTTClient.Connect(strDeviceID, "", "");
      
      if (MQTT_Connect_Result == 1)
      {
        // connected  
        MQTTConnectingSuccessful = true;

        Serial.println("MQTT connected.");        
      }
      else
      {
        // not connected
        Serial.println("MQTT not connected.");
      }
   }

   return MQTTConnectingSuccessful;
}



// Publish a MQTT message with SensorDate (here: 1 Byte, value 0..99) to MQTT Broker
void PublishSensorDate(String SensorName, byte SensorDate)
{

    // take the config strings out of non volatile storage 
    String strKey = preferences.getString("Payload", "");
    String strDeviceID = preferences.getString("DeviceID", "");
    String strSensorID = preferences.getString("SensorID", "");

    // strKey = Payload format string
    // replace the placeholder in this string by DeviceID and SensorID (configurated by the user)
    strKey.replace("$D$", strDeviceID);
    strKey.replace("$S$", strSensorID);
    strKey.replace("$V$", String(SensorDate));
    strKey.replace("$N$", SensorName);
    strKey.replace("#", String(char('\"')));
    
    String MString = strKey;

    int Slength = MString.length();

    // copy the bytes of the payload string in the payload buffer
    for (int j = 0; j < Slength; j++)
    {
        myMQTTClient.PayloadBuffer[j] = MString[j];
    }

    //  take Topic format strong out of non-volatile storage
    String strTopic = preferences.getString("Topic", "");

    // replace the placeholder in this string by DeviceID and SensorID (configurated by the user)
    strTopic.replace("$D$", strDeviceID);
    strTopic.replace("$S$", strSensorID);
    strTopic.replace("$N$", SensorName);

    String SensorTopic = strTopic;

    Serial.println(SensorTopic);

    // Publish the message with topic and payload
    myMQTTClient.Publish(SensorTopic, myMQTTClient.PayloadBuffer, Slength);

    // update state variables to be shown in the GUI
    MQTTClient_LastCommand = MQTTPUBLISH;
    MQTTClient_PublishCount++;

}



// Get and send out sensor data via MQTT
// light data is sampled via ADC 
void GetAndSendSensorData()
{
 
    // read the voltage at ADC-pin, 0..4096, divide by 64, means 0..63
    int analog_value = analogRead(PIN_SENSOR) >> 6;

    // Send sensor data via MQTT
    PublishSensorDate("light", analog_value); 

    Serial.print("VAL:");
    Serial.println(analog_value); 

    // update state variable to be shown in the GUI
    SENSOR_LastValue = analog_value;
}



void loop() {

   // check if request from client was received (PC or smartphone browser with HTML config form)
   int ConfigRequestReceived = GetConfigRequestAndSendHTMLPage();
   if (ConfigRequestReceived == 1)  // yes
   {
      RouterNetworkDeviceState = NETWORKSTATE_LOGINREQUESTED;    // set state variable that a new login attempt follows
      CounterForLoginRouterNetworkAttempt = 40;                  // immediately
   }
    
   if (RouterNetworkDeviceState != NETWORKSTATE_LOGGED)    // if we are not logged in or login attempt requested by user
   {
      if (CounterForLoginRouterNetworkAttempt > 40)        // counter for a delay between login attempts (approx. every 2 s)
      {
        
        GUI_Update();                             // update GUI to show login attempt
        
        LoginRouterNetwork();                     // login attempt
        CounterForLoginRouterNetworkAttempt = 0;
      }
      CounterForLoginRouterNetworkAttempt++;
   }

   GUI_Heartbeat();  // show a heartbeat sign at the GUI
   
   if (RouterNetworkDeviceState == NETWORKSTATE_LOGGED)    // if we are logged in
   {

     CounterForMQTT++;
     if (CounterForMQTT > 40)                              // do following approx every 2 seconds
     {
       CounterForMQTT = 0;

       CounterForPing++;
       if (CounterForPing > 4)                             // do following approx every 10 seconds
       {
          CounterForPing = 0;

          MQTTClient_LastCommand = MQTTPINGREQ;            // update GUI to show ping
          GUI_Update();

          int PingResult = myMQTTClient.Ping();            // Send out Ping to MQTT broker
          if (PingResult == 0)                             // no Ping response
          {
            MQTTClient_Connected = false;                  // set state variable that we need a new MQTT connection
          }
          
       }
              
       if (MQTTClient_Connected == false)                  // no MQTT connection
       {

          MQTTClient_LastCommand = MQTTCONNECT;            // update GUI to connection attempt
          GUI_Update();
          
          int NodeState = ConnectToMQTTBroker();           // try to connect to MQTT Broker

          if (NodeState > 0)                               // MQTT connection established successfully
          {
            MQTTClient_Connected = true;                   // state variable is set that we can go in normal mode and send out sensor data
          }
          else                                             // MQTT connection could not be established
          {
            RouterNetworkDeviceState = NETWORKSTATE_NOTLOGGED;   // we assume that Wifi access was lost
            CounterForLoginRouterNetworkAttempt = 0;             // new Wifi login attempt with 2 s delay
          }
  
       }
       else                  // we are successfully connected with MQTT broker
       {
          GetAndSendSensorData();        
       }

       GUI_Update();         // Update GUI to show new status variables
       
    }
  }
  delay(50);          // wait 50 ms
}
